//___________________________________________________________________
//
//                           Class Console
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/glcanvas.h>
#include <wx/textctrl.h>
//#include <wx/html/helpctrl.h>
#include <wx/help.h>
#include <wx/timer.h>
#include <wx/clipbrd.h>
#include <wx/filename.h>

#include <fstream>

#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "Main.h"

#include "String.h"
#include "File.h"
#include "Pair.h"
#include "Line.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "TextControl.h"

#include "Error.h"
#include "Eval.h"
#include "Init.h"
#include "Stack.h"
#include "Files.h"

#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Mouse.h"
#include "Font.h"
#include "Turtle.h"
#include "Recorder.h"

#include "Event.h"
#include "Console.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"

#include "EdDefsExt.h"
#include "Edit.h"
#include "EdFrame.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(Console, wxTextCtrl)
	EVT_CHAR				(Console::OnChar					)
	EVT_KEY_DOWN		(Console::OnKeyDown				)
	EVT_KEY_UP			(Console::OnKeyUp					)
	EVT_LEFT_DOWN		(Console::OnMouseLeftDown		)
	EVT_RIGHT_DOWN		(Console::OnMouseRightDown		)
	EVT_MIDDLE_DOWN	(Console::OnMouseMiddleDown	)
	EVT_LEFT_UP			(Console::OnMouseLeftUp			)
	EVT_RIGHT_UP		(Console::OnMouseRightUp		)
	EVT_MIDDLE_UP		(Console::OnMouseMiddleUp		)
	EVT_LEFT_DCLICK	(Console::OnMouseLeftDClick	)	
	EVT_RIGHT_DCLICK	(Console::OnMouseRightDClick	)
	EVT_MIDDLE_DCLICK	(Console::OnMouseMiddleDClick	)
	EVT_MOTION			(Console::OnMouseMotion			)
	EVT_MOUSE_EVENTS	(Console::OnMouse					)
	EVT_CONTEXT_MENU	(Console::OnContextMenu			)
	EVT_HELP	(wxEVT_HELP, Console::OnHelp				)
	EVT_MENU(ID_OPEN_FILE_AT_CURSOR, Console::OnOpenFileAtCursor)
	EVT_MENU_RANGE(ID_CONSOLE_INSERT, ID_CONSOLE_INSERT+1000, Console::OnInsert)
END_EVENT_TABLE()

Console::Console(wxWindow* parent, wxWindowID id, 
			const wxString& value, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name,
			bool autosave_,
			int caretpos):
wxTextCtrl(parent, id, value, pos, size, style, validator, name),
autosave(autosave_)
{
	writeBufferSize=0x1000;
	readIndex=0;
	charAvailable=lineAvailable=charMode=false;
	
	standout='\001';
	plainMode='\002';
	insertMode=true;
	
	coord=wxPoint(0,0);
	Max.x=0x1000; // hard coded; should be improved;
	Max.y=0x1000; // but seems to be faily big enough at the moment...
	File::interactive=true;

	textColor=0;
	backgroundColor=(unsigned int)(-1);

#ifdef WIN32
	fontSize=10;
#else
	fontSize=14;
#endif
	fontFamily=wxMODERN;
	fontStyle=wxNORMAL;
	fontWeight=wxNORMAL;
	fontName="Courier New";
	underline=false;
	tabSize=80; // in 0.1 millimeter
	setDefaultStyle();
	mouse=graph->mouse;
	mouseEvents=true;
	for (int i=0; i < contextMenuNr; i++)
		contextMenu[i]=0;

	consoleFileName=theApp->consoleFileName;
	if (autosave)
	{	if (wxFileExists(consoleFileName))
		{	LoadFile(consoleFileName);
			SetInsertionPoint(caretpos);
		}
	}
}

Console::~Console() 
{
	if (contextMenu[0])
		delete contextMenu[0];

	if (autosave)
	{	SaveFile(consoleFileName);
	}
}

void Console::OnChar(wxKeyEvent &event) 
{
	int c=event.GetKeyCode();
	Event::fifo << new EventChar(c);	
	if (charMode) 
	{	if (c==WXK_CANCEL)
		{	WriteText('\n');
			theApp->action=CONTROL_PAUSE;
			return;
		}
		if (c==WXK_PAUSE)
		{	WriteText('\n');
			if (event.ControlDown())
				theApp->action=CONTROL_PAUSE;
			else
				theApp->action=PAUSE;
			return;
		}
		ch=char(c);
		charAvailable=true;
		// we must handle Events ourselfes, 
		// because no foreign code is allowed on the stack for the persistence
		return;
	}
	switch(c)
	{
	case '\t':			WriteText('\t');	break;	// tab
	case WXK_CANCEL:	WriteText('\n');	
							theApp->action=CONTROL_PAUSE;	break;
	case WXK_PAUSE:	WriteText('\n');
							if (event.ControlDown())
								theApp->action=CONTROL_PAUSE;
							else
								theApp->action=PAUSE;
							break;
	case WXK_F7:		theApp->action=ON_STEP;	break;
	case WXK_F9:		theApp->action=ON_RUN;	break;
	default:				event.Skip();
	}
}

void Console::OnKeyDown(wxKeyEvent &event) 
{
	int c=event.GetKeyCode();
	Event::fifo << new EventKeyDown(c);
	if (charMode) 
	{	if (c==WXK_PAUSE || c==WXK_CANCEL 
		|| c==WXK_CONTROL || c==WXK_SHIFT || c==WXK_ALT)
		{	event.Skip();
			return;
		}		
		if ((event.ControlDown() 
		|| event.AltDown() 
		|| event.MetaDown()
		||	c > 255)
		&& !event.ShiftDown())
		{	ch=char(255);
			chExt=c;
			charAvailable=true;
			return;
		}
		if (c==WXK_RETURN 
		||  c==WXK_BACK)
		{	ch=c;
			charAvailable=true;
			return;
		}
		event.Skip();
		return;
	}
	if ((c!=WXK_RETURN 
	  && c!=WXK_NUMPAD_ENTER)
	|| event.ControlDown() || event.AltDown() || event.MetaDown())
	{	if (event.ControlDown() && c=='V')
			Paste();
		else
			event.Skip();
	}else
	{	int index=GetInsertionPoint(); // end of Line
		long x,y;
		PositionToXY(index, &x, &y);
		wxString line=GetLineText(y);
#ifdef WIN32
		line[line.Length()-1]='\n';
#else
		line+='\n';
#endif
		index=XYToPosition(0,y+1);
#ifdef WIN32
		if (index > GetLastPosition() || index < 0)
#else
		if (index > GetLastPosition() ||  index < 0
		|| (index==GetLastPosition() && GetRange(index-1, index)!='\n'))
#endif
		{	SetInsertionPoint(GetLastPosition());
			WriteText('\n');
			SetInsertionPoint(GetLastPosition());
		}else
			SetInsertionPoint(index);
		finishLine(line);
	}
}

void Console::OnKeyUp(wxKeyEvent &event) 
{
	int c=event.GetKeyCode();
	switch (theApp->action)
	{
	case EXIT_LOGO:
	case EXIT_DELAYED:	break;
	default:
		Event::fifo << new EventKeyUp(c);
	}
}

void Console::setMouse(wxMouseEvent &event, MouseP mouse) 
{
	int x=event.GetX(), y=event.GetY();

	if (event.LeftUp())
		mouse->buttons&=~1;
	if (event.RightUp())
		mouse->buttons&=~2;
	if (event.MiddleUp())
		mouse->buttons&=~4;

	if (event.LeftDown())
		mouse->buttons|=1;
	if (event.RightDown())
		mouse->buttons|=2;
	if (event.MiddleDown())
		mouse->buttons|=4;

	if (event.ShiftDown())
		mouse->buttons|=8;
	else
		mouse->buttons&=~8;

	if (event.ControlDown())
		mouse->buttons|=16;
	else
		mouse->buttons&=~16;

	if (event.AltDown())
		mouse->buttons|=32;





	else
		mouse->buttons&=~32;

	if (event.MetaDown())
		mouse->buttons|=64;
	else
		mouse->buttons&=~64;

	HitTest(wxPoint(x,y), (long*)&mouse->textx, (long*)&mouse->texty);
	if (mouseEvents)
		event.Skip();
}

void Console::OnMouse(wxMouseEvent &event) 
{
	setMouse(event, mouse);
}

#define EVENT_ON_MOUSE(NAME)								\
void Console::OnMouse##NAME(wxMouseEvent &event)	\
{																	\
	setMouse(event, mouse);									\
	Event::fifo << new EventTextMouse##NAME(mouse);	\
}

EVENT_ON_MOUSE(LeftDown)
EVENT_ON_MOUSE(RightDown)
EVENT_ON_MOUSE(MiddleDown)

EVENT_ON_MOUSE(LeftUp)
EVENT_ON_MOUSE(RightUp)
EVENT_ON_MOUSE(MiddleUp)

EVENT_ON_MOUSE(LeftDClick)
EVENT_ON_MOUSE(RightDClick)
EVENT_ON_MOUSE(MiddleDClick)

EVENT_ON_MOUSE(Motion)

void Console::updateGraph()
{
	graph->updateGraph();
}

#ifdef WIN32
	#define ACTION					\
		int ibx,icx,idi,isi;		\
		__asm{ mov ibx,ebx }		\
		__asm{ mov icx,ecx }		\
		__asm{ mov idi,edi }		\
		__asm{ mov isi,esi }		\
	   theApp->doAction();		\
		__asm{ mov ebx,ibx }		\
		__asm{ mov ecx,icx }		\
		__asm{ mov edi,idi }		\
		__asm{ mov esi,isi }
#else
	#define ACTION								\
		int ibx,icx,idi,isi;					\
	   asm("mov %%ebx,%0":"=m"(ibx));	\
	   asm("mov %%ecx,%0":"=m"(icx));	\
	   asm("mov %%edi,%0":"=m"(idi));	\
	   asm("mov %%esi,%0":"=m"(isi));	\
	   theApp->doAction();					\
	   asm("mov %0,%%ebx": :"m"(ibx));\
	   asm("mov %0,%%ecx": :"m"(icx));\
	   asm("mov %0,%%edi": :"m"(idi));\
	   asm("mov %0,%%esi": :"m"(isi));
#endif

#ifdef WIN32
	#pragma optimize("", off)
#endif
void Console::dispatchMessages()
{
	while (theApp->Pending())
		theApp->Dispatch();
	theApp->ProcessIdle();
	ACTION;
}
#ifdef WIN32
	#pragma optimize("", on)
#endif

inline bool Console::checkBreak()
{
	static int count=0;
	count++;
	count&=0xfff;	// empirical value
	if (!count)
		dispatchMessages();
	if (theApp->action==CONTROL_PAUSE) 
	{	theApp->action=NOP;
		error(STOP_ERROR);
		return true;
	}
	if (theApp->action==PAUSE)
	{	theApp->action=NOP;
		File::toPending=0;
		lpause(argStack);
	}
	return false;
}

#ifdef WIN32
	#pragma optimize("", off)
#endif
bool Console::kbhit()
{
	bool oldCharMode=charMode;
	charMode=true;
	while (theApp->Pending())
	{	theApp->Dispatch();
		if (charAvailable)
			break;
		ACTION;
		if (theApp->action==EXIT_LOGO)
			return false;
	}
	theApp->ProcessIdle();
	charMode=oldCharMode;
	return charAvailable;
}
#ifdef WIN32
	#pragma optimize("", on)
#endif

char Console::getch()
{
	updateGraph();

	if (!charAvailable)
		while (true)
		{	if (theApp->Pending())
				theApp->Dispatch();
			theApp->ProcessIdle();
			if (charAvailable)
				break;
			if (theApp->action==EXIT_LOGO 
			||  theApp->action==EXIT_DELAYED)
				break;
		}
	charAvailable=false;
	return ch;
}

int Console::getchExt()
{
	charAvailable=false;
	return chExt;
}

#ifdef WIN32
	#pragma optimize("", off)
#endif
char Console::getc_()
{
	if (!lineAvailable) 
	{	updateGraph();

		while (true)
		{
			theApp->Dispatch();
			theApp->ProcessIdle();
			ACTION;

			switch (theApp->action)
			{
			case EXIT_LOGO:
			case EXIT_DELAYED:
									return ' ';
			case ON_RUN:	{	theApp->action=NOP;
									wxCommandEvent cev;
									theApp->edFrame->OnRun(cev); break;
								}
			case ON_STEP:	{	theApp->action=NOP;
									wxCommandEvent cev;
									theApp->edFrame->OnStep(cev); break;
								}
			case PAUSE:		{	char c='p'-'a'+1; // control-p
									theApp->action=NOP;
									return c;
								}
			case CONTROL_PAUSE:
								{	char c='q'-'a'+1; // control-q
									theApp->action=NOP;
									if (graph->recorder != graph->recorderSave)
									{	delete graph->recorder;
										graph->recorder=graph->recorderSave;
									}
									if (graph->turtle->polyMode)
										 graph->turtle->polyEnd();
									if (graph->turtle->tessMode)
										 graph->turtle->tessEnd();
									if (graph->turtle->surfaceMode)
										 graph->turtle->surfaceEnd();

									EventChar   ::function=0;
									EventKeyDown::function=0;
									EventKeyUp  ::function=0;

									EventTextMouseLeftDown  ::function=0;
									EventTextMouseLeftUp    ::function=0;
									EventTextMouseLeftDClick::function=0;

									EventTextMouseRightDown  ::function=0;
									EventTextMouseRightUp    ::function=0;
									EventTextMouseRightDClick::function=0;

									EventTextMouseMiddleDown  ::function=0;
									EventTextMouseMiddleUp    ::function=0;
									EventTextMouseMiddleDClick::function=0;

									EventTextMouseMotion::function=0;

									EventMouseLeftDown  ::function=0;
									EventMouseLeftUp    ::function=0;
									EventMouseLeftDClick::function=0;

									EventMouseRightDown  ::function=0;
									EventMouseRightUp    ::function=0;
									EventMouseRightDClick::function=0;

									EventMouseMiddleDown  ::function=0;
									EventMouseMiddleUp    ::function=0;
									EventMouseMiddleDClick::function=0;

									EventMouseMotion::function=0;

									charMode=false;
									return c;
								}
			default: 
				;
			}
			if (lineAvailable) break;
		}
	}

	char c=readLine[readIndex];
	readIndex++;
	if (c=='q'-'a'+1 || c=='p'-'a'+1 || c=='\n') // control-q, control-p
	{	lineAvailable=false;
		readIndex=0;
//		readLine.resize(0);
		charAvailable=false;
		if (readLine=="reset\n")
			readLine=oldReadLine;
	}
	return c;
}
#ifdef WIN32
	#pragma optimize("", on)
#endif

char Console::read_getc(FILEP strm)
{
	char c=0;
	if (strm==stdin) 
	{	c=getc_();
		if (File::interactive)
		{	switch(c)
			{case 'q'-'a'+1: File::toPending=0;		// control-q
					error(STOP_ERROR);
					return '\n';
			 case 'p'-'a'+1: logo_pause(0);	// control-p
					return read_getc(strm);
			}
		}

	}else if (strm) // reading from a file
		c=getc(strm);
	else if (File::rstringlen > 0  &&  --File::rstringlen > 0)
		c = *File::rstringptr++;	// reading from string
	else
		c=EOF;

	return backslash.decode(c);
}

char Console::putc_(char c, FILEP strm)
{
	if (strm==stdout || strm==stderr) 
	{	switch(c)
		{
		case '\007':theApp->Tone(0, 100); break;

		case '\n':
#ifdef WIN32
						putc_('\r', strm);	// and do the rest...
#endif
		default:		writeBuffer+=char(c);
						if (char(c)=='\n' 
						|| writeBuffer.length() > writeBufferSize)
							flush();
		}
		return 0;
	}
	if (strm==File::textcontrol)
	{	if (File::textControl && (File::textControl->textControl!=0))
		{	TextControl *tc=File::textControl->textControl;
			switch(c)
			{
			case '\007':theApp->Tone(0, 100); break;
			case '\n':
#ifdef WIN32
							putc_('\r', strm);	// and do the rest...
#endif
			default:		tc->writeBuffer+=char(c);
							if (char(c)=='\n' 
								|| tc->writeBuffer.length() > tc->writeBufferSize)
								tc->flush();
			}
		}
		else
		{	File::Writer=File::Stdout;
			putc_(ch, stdout);
		}
		return 0;
	}
	if (c==10)
		return putc(c, strm);
	return putc(c, strm);
}

#ifdef WIN32
	#pragma optimize("", off)
#endif
void Console::flush()
{
	long from, to;
	GetSelection(&from, &to);
#ifdef WIN32
	if (!insertMode)
		to=from+writeBuffer.length();
	Replace(from, to, writeBuffer);
#else
	long ip=GetInsertionPoint();
	if (from < 0)
		from=ip;
	if (to < 0)
		to=ip;
	if (!insertMode)
	{	to=from+writeBuffer.length();
		if (to > GetLastPosition())
			to=GetLastPosition();
//		Remove(from, to);
	}
	Replace(from, to, writeBuffer);
//	WriteText(writeBuffer);
//	long ipnew=ip+writeBuffer.length();
//	SetInsertionPoint(ipnew);
#endif

	while (theApp->Pending())
		theApp->Dispatch();
	theApp->ProcessIdle();

	writeBuffer="";
	ACTION;
}
#ifdef WIN32
	#pragma optimize("", on)
#endif

void Console::setTextSelection(int x1, int y1, int x2, int y2)
{
	int pos1=XYToPosition(x1, y1),

		pos2=XYToPosition(x2, y2);
	SetSelection(pos1, pos2);
}

void Console::update_coords(char ch) 
{
	switch(ch)
	{
	case '\n':
		{	coord.x=0;		// newline
			coord.y++;
			for (int i=0; i < File::marginX; i++) 
				print(stdout,' ');
			break;
		}		
	case '\b': if (coord.x > 0) --coord.x;		// backspace
			break;

	case '\t':	coord.x &=~07; coord.x+=8; break;	// tab
	case '\001':                    // bold mode

	case '\002': break;      // plain mode
	case '\007':               // ...???
		coord.x++;

		if (coord.x > Max.x) 
		{

			if (Max.x != 0) // well, it could be...
			{
				coord.y += coord.x/Max.x;
				coord.x %= Max.x;
			}
		}
		if (coord.y > Max.y) 
			coord.y = Max.y;
	}
}

void Console::print(FILEP strm, char ch) 
{
	if (strm) 
	{	if (File::interactive  &&  strm==stdout) 
		{	//switch (ch)
			//{case '\001': modeBold();	break;
			//case '\002': modePlain();	break;
			//default:
				putc_(ch, strm);
			//}
		}
		else if (strm==File::textcontrol)
		{	if (File::textControl && (File::textControl->textControl!=0))
				putc_(ch, strm);
			else
			{	File::Writer=File::Stdout;
				putc_(ch, stdout);
			}
		}
		else	    // printing to stream but not screen
			putc_(ch, strm);

		if (dribbler)
			putc_(ch, File::Dribbler->file());
	} 
	else if (ch  &&  File::stringlen > 0  &&  --File::stringlen >= 0)
		*File::stringptr++= ch;     // printing to string
}

void Console::finishLine(wxString line)
{
	readLine=line;
	lineAvailable=true;
	readIndex=0;
}

void Console::OnEditRepeat(wxCommandEvent &event)
{
	lineAvailable=true;
	readIndex=0;
}

void Console::OnEditReset(wxCommandEvent &event)
{
	oldReadLine=readLine;
	readLine="reset\n";
	lineAvailable=true;
	readIndex=0;
}

NodeP Console::_Cursor() 
{
	long pos=GetInsertionPoint();
	long x,y;
	PositionToXY(pos, &x, &y);
	return new Pair
	(	new Int(x),
		new Pair(new Int(y),0) );
}

NodeP Console::_setCursor(PairP args) 
{
	PairP vec = args->a_pos_int_vector(1);
	if (!throwing) 
	{	flush();
		wxPoint oldCoord=coord;
		coord.x=IntP(vec-> first())->int_;
		coord.y=IntP(vec->second())->int_;

		if ((coord.x < 0 || coord.y < 0) && !throwing) 
		{	error(BAD_DATA, vec);
			coord=oldCoord;
			return Unbound;
		}
		if (coord.y >= GetNumberOfLines())
		{	int end=GetLastPosition();
			SetInsertionPoint(end);
			long x,y;
			PositionToXY(end, &x, &y);
			int missingLines=coord.y-y;
			wxString CRLFs;
			while (missingLines)
			{
#ifdef WIN32
				CRLFs+="\r\n";
#else
				CRLFs+="\n";
#endif
				missingLines--;
			}
			*this << CRLFs;
		}
		if (coord.x > GetLineLength(coord.y))
		{	SetInsertionPoint(
				XYToPosition( GetLineLength(coord.y), coord.y ));
			WriteText(wxString(
				coord.x+1-GetLineLength(coord.y), ' ').c_str());
		}
		SetInsertionPoint(XYToPosition(coord.x, coord.y));
	}
	return Unbound;
}

NodeP Console::_setTextColor(PairPP arg) 
{
	textColor = arg[1]->a_color(1);

	if (!throwing) 
	{
		backgroundColor = arg[2]->a_color(2);

		if (!throwing) 
		{	flush();
			SetDefaultStyle(wxTextAttr(
				Pen::realcolor(textColor), 
				Pen::realcolor(backgroundColor)));
		}
	}
	return Unbound;
}

void Console::charmodeOn()
{
	if (charMode) 
		return; 
	charMode=true; 
	flush();
}
void Console::charmodeOff() 
{
	if (!EventChar::function)
		charMode=false;
}

void Console::setDefaultStyle()
{
#ifdef WIN32
	wxTextAttr textattr(
		Pen::realcolor(textColor), Pen::realcolor(backgroundColor),
		wxFont(fontSize, fontFamily, fontStyle, fontWeight, underline,
			fontName));
	wxArrayInt tabs;
	tabs.Add(0,10);
	for (int i=0; i<10; i++)
		tabs[i]=tabSize*i;
	textattr.SetTabs(tabs);

	SetDefaultStyle(textattr);
#else
	SetFont(	wxFont(fontSize, fontFamily, fontStyle,
					fontWeight, underline, fontName));
#endif
}

void Console::modeBold     (){fontWeight=wxBOLD;  setDefaultStyle();}
void Console::modeItalic   (){fontStyle=wxITALIC; setDefaultStyle();}
void Console::modeUnderline(){underline=true;     setDefaultStyle();}
void Console::modePlain()
{	fontStyle=wxNORMAL; fontWeight=wxNORMAL; underline=false;
	setDefaultStyle();
}

NodeP Console::_setTextSize(PairP arg)
{
	double s=arg->a_float(1);
	if throwing return Unbound;
	fontSize=int(s);
	setDefaultStyle();
	return Unbound;
}

NodeP Console::_setTextFont(PairP arg)
{
	StringP name=arg->a_string(1);
	if (!throwing)
	{	if (name->length() > 32)
		{	error(BAD_DATA,arg);
			return Unbound;
		}

		flush();
		name=String::make(name->ptr(), 0, 
			name->length(), STRING, String::copy);
		fontName=name->ptr();
		
		setDefaultStyle();
	}
	return Unbound;
}

void Console::toggleInsertMode()
{
	flush();
	insertMode=!insertMode;
#ifdef _WIN32
	keybd_event( VK_INSERT, 0x0, KEYEVENTF_EXTENDEDKEY , 0 );
	keybd_event( VK_INSERT, 0x0, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
#else
	EmulateKeyPress(wxKeyEvent(WXK_INSERT));
//	error(NOT_YET_SUPPORTED, new String("Console::toggleInsertMode"));
#endif
	dispatchMessages();
}

NodeP Console::_insertMode()
{
	if (!insertMode)
		toggleInsertMode();
	return Unbound;
}

NodeP Console::_overwriteMode()
{
	if (insertMode)
		toggleInsertMode();
	return Unbound;
}

NodeP Console::_clearText(PairP)
{
	Clear();
	return Unbound;
}

NodeP Console::_setMargins(PairP args) 
{
	PairP arg;
	arg = args->a_pos_int_vector(1);
	if (!throwing) 
	{	File::marginX = IntP(arg->   first())->int_;
		File::marginY = IntP(arg->second())->int_;
		_clearText(NIL);
	}
	return Unbound;
}

NodeP Console::_standout(NodeP arg) 
{
	char textbuf[65536];
	File::stringptr=textbuf;
	File::stringlen=65535;

	stringbuffer << '\001' << arg << '\002';
	*File::stringptr='\0';
	return String::make(textbuf, 0, strlen(textbuf), STRING, String::copy);
}

void Console::OnSaveDocument(wxString pathname)
{
	NodeP pn=new String(pathname);
	PairPP asp=argStack;
	argStack << pn;
	lsave(asp);
	argStack.pop_till(asp);
	return;
}

char Console::charUnderCursor()
{
	int ip=GetInsertionPoint();
	if (ip==GetLastPosition())
		return '\0';
	wxString tc=GetRange(ip, ip+1);
	return tc[0];
}

wxString Console::wordUnderCursor() 
{
	long selstart,selend;
	GetSelection(&selstart, &selend);
	wxString line;

	if (selstart==selend)
	{	long x,y;
		if (selstart < 0)
			selstart=GetInsertionPoint();
		PositionToXY(selstart, &x, &y);
		line=GetLineText(y);
		long lineindex=XYToPosition(0,y);

		int pos=selstart-lineindex,
			start=pos-1,
			len=line.length();

		if (start<0)
			start=0;
		else
		{	while (start>=0 && isalnum(line[(unsigned int)start])) {start--;}
			start++;
		}
		int end=pos;
		while (isalnum(line[(unsigned int)end]) && end<len) {end++;}

		line=line.Mid(start, end-start);
	}else
	{	line=GetStringSelection();
	}
	return line;
}

void Console::OnHelp(wxHelpEvent &event) 
{
	wxString line=wordUnderCursor();
#ifdef WIN32
	theApp->help->KeywordSearch(line);
#else
	theApp->help->DisplayLow(line);
#endif
}

void Console::OnOpenFileAtCursor(wxCommandEvent &event)
{
	wxString word=wordUnderCursor()+".lg";
	if (wxFileExists(word))
		theApp->edFrame->FileOpen(word);
	else
	{	wxFileName fn(environment.logolib, word);
		if (fn.FileExists())
			theApp->edFrame->FileOpen(word);
	}
}


bool File::checkBreak() {return console->checkBreak();}

char File::    getch() {return console->getch();}
int  File:: getchExt() {return console->getchExt();}

#ifdef WIN32
	#pragma optimize("", off)
#endif
bool File::    kbhit() {return console->kbhit();}
#ifdef WIN32
	#pragma optimize("", on)
#endif


char File::read_getc() 
{
	char c=console->read_getc(file());
	if (c=='\n')
	{
		Line::readnr++;
//		cout << "readnr=" << Line::readnr << endL;
	}
	return c;
}
void File::print(FILEP strm, char ch) {console->print(strm, ch);}

void File::charmodeOn() {console->charmodeOn();}
void File::charmodeOff() {console->charmodeOff();}

NodeP lCursor(PairPP)     {return console->_Cursor();}
NodeP lsetCursor(PairPP arg) {return console->_setCursor(arg[1]);}
NodeP lsetTextColor(PairPP arg) {return console->_setTextColor(arg);}
NodeP lsetTextSize(PairPP arg) {return console->_setTextSize(arg[1]);}
NodeP lsetTextFont(PairPP arg) {return console->_setTextFont(arg[1]);}
NodeP linsertMode(PairPP)     {return console->_insertMode();}
NodeP loverwriteMode(PairPP)     {return console->_overwriteMode();}
NodeP lboldTextMode(PairPP)     {console->modeBold(); return Unbound;}
NodeP litalicTextMode(PairPP)     {console->modeItalic(); return Unbound;}
NodeP lunderlineTextMode(PairPP)    {console->modeUnderline(); return Unbound;}
NodeP lplainTextMode(PairPP)     {console->modePlain(); return Unbound;}
NodeP lclearText(PairPP)     {return console->_clearText(NIL);}
NodeP lsetMargins(PairPP arg) {return console->_setMargins(arg[1]);}
NodeP lstandout(PairPP arg) {return console->_standout(arg[1]);}

NodeP ldispatchMessages(PairPP arg) 
{
	console->updateGraph();	
	console->dispatchMessages(); 
	return Unbound;
}

NodeP lCharUnderCursor(PairPP) 
{
	char c[2];
	c[0]=console->charUnderCursor();
	c[1]=0;
	return new String(c);
}

NodeP lWordUnderCursor(PairPP) 
{
	wxString word=console->wordUnderCursor();
	return new String(word);
}

NodeP lsetTextSelection(PairPP arg)
{
	PairP start=arg[1]->a_pos_int_vector(1),
		end=arg[2]->a_pos_int_vector(2);
	if throwing return Unbound;

	int startx=int(start->first()->getnumber()),
		starty=int(start->second()->getnumber()),
		endx=int(end->first()->getnumber()),
		endy=int(end->second()->getnumber());

	console->setTextSelection(startx, starty, endx, endy);
	return Unbound;
}

NodeP lenableTextMouseEvents(PairPP)
{
	console->mouseEvents=true;
	return Unbound;
}

NodeP ldisableTextMouseEvents(PairPP)
{
	console->mouseEvents=false;
	return Unbound;
}

Console::MenuTree Console::command[]=
{
#include "toc.h"
	{0,0}
};

void Console::createMenu(int &used, int &i, int level)
{
	int parent=used;
	contextMenu[parent] = new wxMenu;

	while (true)
	{	int newlevel=command[i].level;

		if (newlevel >= level)
		{	
			int nextlevel=command[i+1].level;
			if (nextlevel == newlevel)
			{	contextMenu[parent]->Append(ID_CONSOLE_INSERT+i, command[i].str);
				i++;
			}
			else if (nextlevel > newlevel)
			{	used++;
				int item=i, child=used;
				i++;
				createMenu(used, i, newlevel);
				contextMenu[parent]->Append(ID_CONSOLE_INSERT+item, 
					command[item].str, contextMenu[child]);
				nextlevel=command[i].level;
				if (nextlevel < newlevel)
					break;
			}
			else
			{	contextMenu[parent]->Append(ID_CONSOLE_INSERT+i, command[i].str);
				i++;
				break;
			}
		}
		else
		{	i++;
			break;
		}
	};
}

void Console::OnContextMenu(wxContextMenuEvent &event)
{
	if (!contextMenu[0])
	{
		int used=0, i=0, level=1; 
		createMenu(used, i, level);

		wxMenu &m = *contextMenu[0];
		m.AppendSeparator();
		m.Append(wxID_UNDO, _("&Undo"));
		m.Append(wxID_REDO, _("&Redo"));
		m.AppendSeparator();
		m.Append(wxID_CUT, _("Cu&t"));
		m.Append(wxID_COPY, _("&Copy"));
		m.Append(wxID_PASTE, _("&Paste"));
		m.Append(wxID_CLEAR, _("&Delete"));
		m.AppendSeparator();
		m.Append(wxID_SELECTALL, _("Select &All"));
		m.AppendSeparator();
		m.Append(ID_OPEN_FILE_AT_CURSOR, _("&Open File at Cursor"));
	}
	PopupMenu(contextMenu[0]);

	return;
}

void Console::OnInsert(wxCommandEvent &event)
{
	WriteText(command[event.GetId()-ID_CONSOLE_INSERT].str);
	WriteText(' ');
}

void Console::Paste()
{
	wxClipboard clipboard;
	if (clipboard.Open())
	{
		if (clipboard.IsSupported(wxDF_TEXT))
		{
			wxTextDataObject data;
			clipboard.GetData(data);
			WriteText(data.GetText());
		}  
		clipboard.Close();
	}
}

NodeP lConsoleSetFocus(PairPP)
{
	console->SetFocus();
	return Unbound;
}
