//___________________________________________________________________
//
//                           Class Point
//___________________________________________________________________

class wxTextInputStream;
class wxTextOutputStream;
class wxDataInputStream;
class wxDataOutputStream;

class Point
{	
	public:
		union {double x; double a[1];};
		double y,z;

		Point(){x=y=z=0;}
		Point(double r){x=r; y=r; z=r;}
		Point(const double &_x, const double &_y, const double &_z){x=_x; y=_y; z=_z;}

		Point operator+(){return *this;}
		Point operator-(){return Point(-x,-y,-z);}

		Point round(){return Point(int(x+0.5), int(y+0.5), int(z+0.5));}
		Point   abs(){return Point(Float::abs(x), Float::abs(y), Float::abs(z));}
		Point  sqrt(){return Point(::sqrt(x), ::sqrt(y), ::sqrt(z));}

		Point operator+(const double &b){return Point(x+b, y+b, z+b);}
		Point operator-(const double &b){return Point(x-b, y-b, z-b);}
		Point operator*(const double &b){return Point(x*b, y*b, z*b);}
		Point operator/(const double &b){return Point(x/b, y/b, z/b);}
		Point operator%(const double &b){return Point(
			Float::mod(x, b), Float::mod(y, b), Float::mod(z, b));}

		Point operator+(const Point &b){return Point(x+b.x, y+b.y, z+b.z);}
		Point operator-(const Point &b){return Point(x-b.x, y-b.y, z-b.z);}
		Point operator*(const Point &b){return Point(x*b.x, y*b.y, z*b.z);}
		Point operator/(const Point &b){return Point(x/b.x, y/b.y, z/b.z);}
		Point operator%(const Point &b){return Point(
			Float::mod(x, b.x), Float::mod(y, b.y), Float::mod(z, b.z));}

		Point &operator+=(const double &b){x+=b; y+=b; z+=b; return *this;}
		Point &operator-=(const double &b){x-=b; y-=b; z-=b; return *this;}
		Point &operator*=(const double &b){x*=b; y*=b; z*=b; return *this;}
		Point &operator/=(const double &b){x/=b; y/=b; z/=b; return *this;}

		Point &operator=(const Point &b)
		{	x=b.x; y=b.y; z=b.z; 
			return *this;
		}
		Point &operator+=(const Point &b){x+=b.x; y+=b.y; z+=b.z; return *this;}
		Point &operator-=(const Point &b){x-=b.x; y-=b.y; z-=b.z; return *this;}
		Point &operator*=(const Point &b){x*=b.x; y*=b.y; z*=b.z; return *this;}
		Point &operator/=(const Point &b){x/=b.x; y/=b.y; z/=b.z; return *this;}

		bool operator==(const double &b){return (x==b && y==b && z==b);}
		bool operator!=(const double &b){return (x!=b || y!=b ||z!=b);}
		bool operator< (const double &b){return (Norm() < b);}
		bool operator> (const double &b){return (Norm() > b);}

		bool operator==(const Point &b){return (x==b.x && y==b.y && z==b.z);}
		bool operator!=(const Point &b){return (x!=b.x || y!=b.y || z!=b.z);}
		bool operator< (const Point &b){return Norm() < b.Norm();}
		bool operator> (const Point &b){return Norm() > b.Norm();}

		double Norm() const {return ::sqrt(x*x+y*y+z*z);}
		double min_() const {return x<y? (x<z?x:z):(y<z?y:z);}
		bool in(const Point &a, const Point &b) 
		{return (a.x<=x && x<=b.x && a.y<=y && y<=b.y && a.z<=z && z<=b.z);}

		Point X(const Point &b)
		{	return Point(y*b.z-z*b.y, z*b.x-x*b.z, x*b.y-y*b.x);
		}

	friend wxTextInputStream &operator>>(wxTextInputStream &istr, Point &p);
	friend wxTextOutputStream &operator<<(wxTextOutputStream &ostr, Point &p);
	friend wxDataInputStream &operator>>(wxDataInputStream &istr, Point &p);
	friend wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Point &p);
};

typedef Point* PointP;

class PointFiFo		// is a First-in-First-out buffer class for Points.
{
	public:
		bool allocated, _overflow;
		int size, expand_size;
		PointP base, max, in, out;

		PointFiFo(int expand_size_) 
		{	expand_size=expand_size_;
			allocated=_overflow=false; 
			size=0; 
			base=max=in=out=0; 
			expand();
		}
		~PointFiFo();

		PointFiFo &operator<<(const Point val);
		PointFiFo &operator>>(Point &val);

		operator bool() {return in!=out;}
		PointFiFo &operator=(PointFiFo &b);
		bool overflow() {return _overflow;}
		void reset_in() {in=base;}
		void reset_out() {out=base;}
		void reset() {in=out=base;}
		void expand();
};

class Matrix
{
	public:
		Point x,y,z;

		Matrix() {}
		Matrix(double r) {x=Point(r); y=Point(r); z=Point(r);}
		Matrix(const Point &_x, const Point &_y, const Point &_z)
		{	x=_x; y=_y; z=_z;
		}

		Matrix &operator=(const Matrix &b)
		{	x=b.x; y=b.y; z=b.z; 
			return *this;
		}

		Point operator*(const Point &b){return x*b.x+y*b.y+z*b.z;}

		Matrix invert()
		{	double 
				d1=y.y*z.z-z.y*y.z,
				d2=y.z*z.x-z.z*y.x,
				d3=y.x*z.y-z.x*y.y,
	
				det=x.x*d1+x.y*d2+x.z*d3;

			if (det==0)
				return *this;

			Matrix b;
			b.x.x=d1/det;
			b.x.y=(z.y*x.z-x.y*z.z)/det;
			b.x.z=(x.y*y.z-y.y*x.z)/det;				
			b.y.x=d2/det;
			b.y.y=(z.z*x.x-x.z*z.x)/det;
			b.y.z=(x.z*y.x-y.z*x.x)/det;				
			b.z.x=d3/det;
			b.z.y=(z.x*x.y-x.x*z.y)/det;
			b.z.z=(x.x*y.y-y.x*x.y)/det;
			return b;
		}

	friend wxTextInputStream &operator>>(wxTextInputStream &istr, Matrix &m);
	friend wxTextOutputStream &operator<<(wxTextOutputStream &ostr, Matrix &m);
	friend wxDataInputStream &operator>>(wxDataInputStream &istr, Matrix &m);
	friend wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Matrix &m);
};
