//___________________________________________________________________
//
//                         Class Recorder
//___________________________________________________________________

class Graph;
class Screen_;
class Font_;
class Turtle;
class PtrToInt;
class IntToPtr;
class Recorder;
class wxTextInputStream;
class wxTextOutputStream;

class Command
{
	friend class CommandList;
	friend class CommandListIterator;
	Command *prev, *next;
	public:
		enum Type	   // types of graphics moves that can be recorded
		{	NOP,		// and some internal help commands to keep things weird
			GRAPHIC, ENDGRAPHIC, 
			MOVETO, LINETO, LINETO3D, 
			SETMATRIX, SETMATRIX4X4, PUSHMATRIX, POPMATRIX,
			SETPIXEL,
			POLYSTART, POLYPOINT, POLYEND, 
			GLBEGIN, TESSPOINT,
			QUADSTRIPSTART, QUADSTRIP, QUADSTRIPEND,
			FILL, FILLRECT, FILLELLIPSE, 
			SPHERE, ELLIPSOID, PARTIALELLIPSOID, FILLPIE,
			PERSPECTIVE, SETEYE, WRAP, MODE,
			SETSCREENCOLOR, MATERIAL, LIGHT, SETPENCOLOR,SETFLOODCOLOR, SETSHADOWCOLOR, 
			SHININESS, 
			SETPENDOWN, 
			SETPOLYSMOOTH, SETLINESMOOTH, SETPOINTSMOOTH,
			SETDEPTHTEST, SETLIGHTING, SETDITHER, SETTEXTURED,
			SETPENRLE, SETCYLINES, SETSHADOWS, SETFOGENABLED, SETPENMODE, 
			SETPENSIZE, SETPENPATTERN, SETDEPTHFUNC, 
			TEXTURE, TESSWINDINGRULE,
			LABEL, 
			SETLIGHTPOS,
			LASTCOMMAND
		};
		Type type;

		static const charP 
			penUpDownString[], commandTypeString[];  // for debugging


		Command() {prev=next=this;}
		virtual ~Command() {}
		virtual void play(Recorder &rec) {}

		virtual void setPtrs() {}
		virtual wxTextInputStream  &read(wxTextInputStream &istr);
		virtual wxTextOutputStream &write(wxTextOutputStream &ostr);
		virtual wxDataInputStream  &read(wxDataInputStream &istr);
		virtual wxDataOutputStream &write(wxDataOutputStream &ostr);
		virtual void writePS(wxTextOutputStream &ostr, Recorder &recorder) {}
//		virtual void writeVRML(wxTextOutputStream &ostr, Recorder &recorder);

	friend wxTextInputStream &operator>>(wxTextInputStream &istr, Command &command) {return command.read(istr);}
	friend wxTextOutputStream &operator<<(wxTextOutputStream &ostr, Command &command) {return command.write(ostr);}
	friend wxDataInputStream &operator>>(wxDataInputStream &istr, Command &command) {return command.read(istr);}
	friend wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Command &command) {return command.write(ostr);}
};

class CommandList: public Command
{
	friend class CommandListIterator;
	public:
		bool empty() {return next==this;}
		Command *valid(Command *command) {return command!=this ?   command  :  0;}
		Command *begin() {return valid(next);}
		Command *end() {return this;}

		Command *takeOut(Command *command)
		{	command=valid(command);
			if (command)
			{	command->next->prev = command->prev;
				command->prev->next = command->next;
			}
			return command;
		}
		void clear()
		{	while (!empty()) 
				delete takeOut(next);
		}
		~CommandList() {clear();}

		CommandList &push_back(Command *command)
		{	Command *here=prev;
			if (!here)
				here=next;
			command->prev=here;
			command->next=here->next;
			here->next->prev=command;
			here->next=command;
			return *this;
		}
};

class CommandListIterator // This might be templated in the future (but I don`t like templates so very much, so I probably won`t do it)
{
	Command *here;
	CommandList *list;
	public:
		CommandListIterator(CommandList *list_): here(0), list(list_) {}
		Command *operator++() {here = list->valid( here ?   here->next  :  list->next ); return here;}
		Command *operator--() {here = list->valid( here ?   here->prev  :  list->prev ); return here;}
		Command *operator++(int) {here = list->valid( here ?   here->next  :  list->next ); return here;}
		Command *operator--(int) {here = list->valid( here ?   here->prev  :  list->prev ); return here;}
		Command *operator=(Command *command) {here=command; return here;}
		Command *operator->() {return here;}
		Command *operator*() {return here;}
		operator unsigned int() {return (unsigned int)(here);}
		bool operator!=(Command *command) {return command!=(Command *)this;}

	friend wxTextInputStream &operator>>(wxTextInputStream &istr, CommandListIterator &i);
	friend wxTextOutputStream &operator<<(wxTextOutputStream &ostr, CommandListIterator &i);
	friend wxDataInputStream &operator>>(wxDataInputStream &istr, CommandListIterator &i);
	friend wxDataOutputStream &operator<<(wxDataOutputStream &ostr, CommandListIterator &i);
};

class Recorder: public CommandList
{
	friend class R_MoveTo;
	public:
		unsigned int refCount;
		bool recording, pendown;
		bool /*volatile*/ inReplay, restartFlag, stop_play, stop_confirm;
		bool inload, insave;
		int action, tabs;
		R_MoveTo *lastMove;
		Graph *graph;
		Screen_ *screen;
		Turtle *turtle;

		static PtrToInt *ptrToInt;
		static IntToPtr *intToPtr;

		static const int tableSize;

		Recorder(Graph *graph_);
		~Recorder();

		void off() {recording=false;}
		void  on() {recording=true;}

		void clear() {lastMove=0; CommandList::clear();}
		void move();

		void playAgain(), stop(), updateView(Turtle &turtle);

		static DWORD WINAPI 
			playCommands(LPVOID lpv);

		wxString rgb(Color c);

		void
			save(const char *name),
			load(const char *name),
			saveText(const char *name),
			loadText(const char *name),
			writePS(const char *name);

		CommandList &operator<<(Command *command)
		{	if (recording)
			{	if (command)
					push_back(command);
				else if (!throwing)
					error(OUT_OF_MEM);
			}
			return *this;
		}

	friend wxTextInputStream &operator >>(wxTextInputStream &istr, Recorder &recorder);
	friend wxTextOutputStream &operator <<(wxTextOutputStream &ostr, Recorder &recorder);
	friend wxDataInputStream &operator >>(wxDataInputStream &istr, Recorder &recorder);
	friend wxDataOutputStream &operator <<(wxDataOutputStream &ostr, Recorder &recorder);
};

class R_Graphic: public Command
{
	Turtle recturtle;
	double scale;
	Recorder *recorder;
	public:
		R_Graphic() {type=GRAPHIC;}
		R_Graphic(Turtle &turtle, Recorder *recorder_, double scale);
		~R_Graphic();
		void play(Recorder &rec);
		void setPtrs();
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_MoveTo: public Command
{
	public:
		Point point;

		R_MoveTo() {type=MOVETO;}
		R_MoveTo(Turtle &turtle, Recorder &recorder);
		R_MoveTo(Turtle &turtle, Recorder &recorder, Point point_);
		R_MoveTo(Turtle &turtle, Recorder &recorder, double x, double y);

		virtual void init(Turtle &turtle, Recorder &recorder, 
			Point point_)
		{	type=MOVETO;
			point=point_;
			recorder.lastMove=this;
		}
		virtual void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_LineTo: public R_MoveTo
{
	Matrix u;
	public:
		R_LineTo() {type=LINETO;}
		R_LineTo(Turtle &turtle, Recorder &recorder);
		R_LineTo(Turtle &turtle, Recorder &recorder, Point point_);
		R_LineTo(Turtle &turtle, Recorder &recorder, double x, double y);

		void init(Turtle &turtle, Recorder &recorder, Point point_)
		{	type=LINETO;
			point=point_;
			u=turtle.u;
			recorder.lastMove=(R_MoveTo*)this;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_LineTo3d: public R_MoveTo
{
	Matrix u;
	public:
		R_LineTo3d() {}
		R_LineTo3d(Turtle &turtle, Recorder &recorder);
		R_LineTo3d(Turtle &turtle, Recorder &recorder, Point point_);
		R_LineTo3d(Turtle &turtle, Recorder &recorder, double x, double y);

		void init(Turtle &turtle, Recorder &recorder, Point point_);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetMatrix: public Command
{
	Point basex,basey,basez,trans;
	public:
		R_SetMatrix() {}
		R_SetMatrix(Point basex_, Point basey_, Point basez_, Point trans_)
		{	type=SETMATRIX; 
			basex=basex_;	
			basey=basey_;
			basez=basez_;
			trans=trans_;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetMatrix4x4: public Command
{
	double m[16];
	public:
		R_SetMatrix4x4() {}
		R_SetMatrix4x4(double m_[16])
		{	type=SETMATRIX4X4; 
			for (int i=0; i<16; i++)
				m[i]=m_[i];
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_PushMatrix: public Command
{
	public:
		R_PushMatrix() 
		{	type=PUSHMATRIX;
		}
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_PopMatrix: public Command
{
	public:
		R_PopMatrix() 
		{	type=POPMATRIX;
		}
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetPixel: public R_MoveTo
{
	Color color;
	public:
		R_SetPixel() {}
		R_SetPixel(Point &p, Color c) 
		{	type=SETPIXEL; point=p; color=c;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_PolyStart: public Command
{
	Point normal, p1,p2,p3, t1,t2,t3;
	public:
		R_PolyStart() {}
		R_PolyStart(Turtle &turtle, Point &normal_, Point &p3_, Point &t3_);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_PolyPoint: public Command
{
	Point p, t;
	public:
		R_PolyPoint() {}
		R_PolyPoint(Point &p_, Point &t_)
		{	type=POLYPOINT; 
			p=p_;
			t=t_;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_PolyEnd: public Command
{
	public:
		R_PolyEnd()
		{	type=POLYEND; 
		}
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_glBegin: public Command
{
	GLenum prim;
	Point normal;
	public:
		R_glBegin() {}
		R_glBegin(GLenum prim_, Point normal_):
			normal(normal_), prim(prim_)
		{	type=GLBEGIN;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_tessPoint: public Command
{
	Point p, t;
	Color c;
	public:
		R_tessPoint() {}
		R_tessPoint(Point &p_, Point &t_, Color c_)
		{	type=TESSPOINT; 
			p=p_;
			t=t_;
			c=c_;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_QuadStripStart: public Command
{
	Point normal,onormal, 
		p1,t1, p2,t2, p3,t3, p4,t4;
	Color c1,c2,c3,c4;
	public:
		R_QuadStripStart() {}
		R_QuadStripStart(Point &normal_,Point &onormal_, 
			Color &c1_, Point &p1_, Point &t1_, 
			Color &c2_, Point &p2_, Point &t2_,
			Color &c3_, Point &p3_, Point &t3_,
			Color &c4_, Point &p4_, Point &t4_)
		{	type=QUADSTRIPSTART; 
			normal=normal_;
			onormal=onormal_;
			c1=c1_; p1=p1_; t1=t1_;
			c2=c2_; p2=p2_; t2=t2_;
			c3=c3_; p3=p3_; t3=t3_;
			c4=c4_; p4=p4_; t4=t4_;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_QuadStrip: public Command
{
	Point normal,onormal, p3,t3, p4,t4;
	Color c3,c4;
	public:
		R_QuadStrip() {}
		R_QuadStrip(Point &normal_, Point &onormal_, 
			Color &c3_, Point &p3_, Point &t3_,
			Color &c4_, Point &p4_, Point &t4_)
		{	type=QUADSTRIP; 
			normal=normal_;
			onormal=onormal_;
			c3=c3_; p3=p3_; t3=t3_;
			c4=c4_; p4=p4_; t4=t4_;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_QuadStripEnd: public Command
{
	public:
		R_QuadStripEnd()
		{	type=QUADSTRIPEND; 
		}
		void play(Recorder &rec);
};

class R_bool: public Command
{
	public:
		bool on;

		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_Fill: public R_bool
{
	public:
		R_Fill() {}
		R_Fill(bool surface_)
		{	type=FILL; 
			on=surface_;
		}
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_FillRect: public Command
{
	Point p1, p2, ori;
	public:
		R_FillRect() {}
		R_FillRect(Turtle &turtle, Point p1_, Point p2_);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_FillEllipse: public Command
{
	Point radius, ori;
	public:
		R_FillEllipse() {}
		R_FillEllipse(Turtle &turtle, Point radius_);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_Sphere: public Command
{
	double radius;
	int slices, stacks;
	Point ori;
	public:
		R_Sphere() {}

		R_Sphere(Turtle &turtle, double radius_, int slices_, int stacks_)
		{	type=SPHERE; 
			radius=radius_;
			slices=slices_;
			stacks=stacks_;
			ori=Point(
				turtle.roll(), 
				turtle.pitch(), 
				turtle.getHeading());
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_Ellipsoid: public Command
{
	Point radius, ori;
	int slices, stacks;
	public:
		R_Ellipsoid() {}
		R_Ellipsoid(Turtle &turtle, Point radius_, int slices_, int stacks_)
		{	type=ELLIPSOID; 
			radius=radius_;
			slices=slices_;
			stacks=stacks_;
			ori=Point(
				turtle.roll(), 
				turtle.pitch(), 
				turtle.getHeading());
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_partialEllipsoid: public Command
{
	Point radius, ori;
	double slicesstart, slicesend, stacksstart, stacksend;
	int slices, stacks;
	public:
		R_partialEllipsoid() {}
		R_partialEllipsoid(Turtle &turtle, Point radius_, 
						double slicesstart_, double slicesend_, int slices_,
						double stacksstart_, double stacksend_, int stacks_)
		{	type=PARTIALELLIPSOID; 
			radius=radius_;
			slicesstart=slicesstart_;
			slicesend=slicesend_;
			slices=slices_;
			stacksstart=stacksstart_;
			stacksend=stacksend_;
			stacks=stacks_;
			ori=Point(
				turtle.roll(), 
				turtle.pitch(), 
				turtle.getHeading());
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_FillPie: public Command
{

	double heading, angle;
	Point radius;
	public:
		R_FillPie() {}
		R_FillPie(Turtle &turtle, double angle_, Point radius_);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_Perspective: public Command
{
	public:
		R_Perspective() {type=PERSPECTIVE;}
		void play(Recorder &rec);
};

class R_SetEye: public Command
{
	Point eye, center, up;
	public:
		R_SetEye() {type=SETEYE;}
		R_SetEye(Point &eye_, Point &center, Point &up);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_wrap: public Command
{
	public:
		R_wrap() {type=WRAP;}
		void play(Recorder &rec);
};

class R_mode: public Command
{
	public:
		Turtle::Mode mode;
		R_mode() {type=MODE;}
		R_mode(Turtle &turtle);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_Color: public Command
{
	public:
		Color color;
		R_Color() {}
		void play(Recorder &rec) {}
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_SetScreenColor: public R_Color
{
	public:
		R_SetScreenColor() {type=SETSCREENCOLOR;}
		R_SetScreenColor(Color _color) {type=SETSCREENCOLOR; color=_color;}
		void play(Recorder &rec);
};

class R_Material: public R_Color
{
		int gltype;
	public:
		R_Material() {type=MATERIAL;}
		R_Material(int _gltype, Color _color) 
		{	type=MATERIAL; gltype=_gltype; color=_color;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_Light: public R_Color
{
		int gltype;
	public:
		R_Light() {type=LIGHT;}
		R_Light(int _gltype, Color _color) 
		{	type=LIGHT; gltype=_gltype; color=_color;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_SetPenColor: public R_Color
{
	public:
		R_SetPenColor() {type=SETPENCOLOR;}
		R_SetPenColor(Turtle &turtle);
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetFloodColor: public R_Color
{
	public:
		R_SetFloodColor() {type=SETFLOODCOLOR;}
		R_SetFloodColor(Turtle &turtle);
		void play(Recorder &rec); 
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetShadowColor: public R_Color
{
	public:
		R_SetShadowColor() {type=SETSHADOWCOLOR;}
		R_SetShadowColor(Turtle &turtle);
		void play(Recorder &rec); 
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_Shininess: public Command
{
	float s;
	public:
		R_Shininess() {type=SHININESS;}
		R_Shininess(float s_) 
		{	type=SHININESS; s=s_;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_SetPenDown: public R_bool
{
	public:
		R_SetPenDown() {type=SETPENDOWN;}
		R_SetPenDown(Turtle &turtle);
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetPolySmooth: public R_bool
{
	public:
		R_SetPolySmooth() {type=SETPOLYSMOOTH;}
		R_SetPolySmooth(bool smooth) {type=SETPOLYSMOOTH; on=smooth;}
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetLineSmooth: public R_bool
{
	public:
		R_SetLineSmooth() {type=SETLINESMOOTH;}
		R_SetLineSmooth(bool smooth) {type=SETLINESMOOTH; on=smooth;}
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetPointSmooth: public R_bool
{
	public:
		R_SetPointSmooth() {type=SETPOINTSMOOTH;}
		R_SetPointSmooth(bool on_) {type=SETPOINTSMOOTH; on=on_;}
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetDepthTest: public R_bool
{
	public:
		R_SetDepthTest() {type=SETDEPTHTEST;}
		R_SetDepthTest(bool on_) {type=SETDEPTHTEST; on=on_;}
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetLighting: public R_bool
{
	public:
		R_SetLighting() {type=SETLIGHTING;}
		R_SetLighting(bool on_) {type=SETLIGHTING; on=on_;}
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetDither: public R_bool
{
	public:
		R_SetDither() {type=SETDITHER;}
		R_SetDither(bool on_) {type=SETDITHER; on=on_;}
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetTextured: public R_bool
{
	public:
		R_SetTextured() {type=SETTEXTURED;}
		R_SetTextured(bool on_) {type=SETTEXTURED; on=on_;}
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetPenRLE: public R_bool	// turtle->pen.roundLineEnds
{
	public:
		R_SetPenRLE() {type=SETPENRLE;}
		R_SetPenRLE(Turtle &turtle);
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetCyLines: public R_bool
{
	public:
		R_SetCyLines() {type=SETCYLINES;}
		R_SetCyLines(Turtle &turtle);
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetShadows: public R_bool
{
	public:
		R_SetShadows() {type=SETSHADOWS;}
		R_SetShadows(Turtle &turtle);
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetFogEnabled: public R_bool
{
	public:
		R_SetFogEnabled() {type=SETFOGENABLED;}
		R_SetFogEnabled(bool fog);
		void play(Recorder &rec);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetPenMode: public Command
{
	Pen::WinMode penMode;
	Pen::Mode internalMode;
	public:
		R_SetPenMode() {type=SETPENMODE;}

		R_SetPenMode(Turtle &turtle);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_SetPenSize: public Command
{
	Point penSize;
	public:
		R_SetPenSize() {type=SETPENSIZE;}
		R_SetPenSize(Turtle &turtle);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetPenPattern: public Command
{
	int penPattern;
	public:
		R_SetPenPattern() {type=SETPENPATTERN;}
		R_SetPenPattern(Turtle &turtle);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_SetDepthFunc: public Command
{
	int func;
	public:
		R_SetDepthFunc() {type=SETDEPTHFUNC;}
		R_SetDepthFunc(int func_);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_Texture: public Command
{
	public:
		bool set;
		int x,y, texi;
		ColorP pixels;
		TextureP tex;

		static int maxnr;
		static PairP textures;

		R_Texture() {type=TEXTURE; pixels=0; set=false;}
		R_Texture(TextureP tex_);
		~R_Texture();
		void play(Recorder &rec);
		void setPtrs();
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_TessWindingRule: public Command
{
	int rule;
	public:
		R_TessWindingRule() {type=TESSWINDINGRULE;}
		R_TessWindingRule(int rule_) {type=TESSWINDINGRULE; rule=rule_;}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

class R_Label: public Command
{
	wxString str;
	Point ori;
	Font_ font;
	public:
		R_Label() {type=LABEL;}
		R_Label(Turtle &turtle, wxString str_);
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};

class R_SetLightPos: public R_MoveTo
{
	public:
		R_SetLightPos() {}
		R_SetLightPos(Point &p) 
		{	type=SETLIGHTPOS; point=p;
		}
		void play(Recorder &rec);
		wxTextInputStream  &read(wxTextInputStream &istr);
		wxTextOutputStream &write(wxTextOutputStream &ostr);
		wxDataInputStream  &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void writePS(wxTextOutputStream &ostr, Recorder &recorder);
};
